@@ -1337,7 +1337,7 @@ void tryhash_word_as_command(const wordunit_T *w)
 {
     if (w != NULL && w->next == NULL && w->wu_type == WT_STRING) {
 	wchar_t *cmdname = unquote(w->wu_string);
-	if (wcschr(cmdname, L'/') == NULL) {
+	if (cmdname != NULL && wcschr(cmdname, L'/') == NULL) {
 	    char *mbsname = malloc_wcstombs(cmdname);
 	    get_command_path(mbsname, false);
 	    free(mbsname);
@@ -2123,7 +2123,10 @@ int array_dump_all(const wchar_t *argv0)
 void array_remove_elements(
 	variable_T *array, size_t count, void *const *indexwcss)
 {
-    long indices[count];
+    size_t extended_count = count;
+    if (extended_count == 0)
+	extended_count = 1; // A variable-length array must ot be empty.
+    long indices[extended_count];
 
     assert((array->v_type & VF_MASK) == VF_ARRAY);
 
